import os
import glob

from mmengine.registry import HOOKS
from mmengine.dist import master_only
from mmengine.hooks.logger_hook import LoggerHook

from typing import Dict, Optional, Union


@HOOKS.register_module()
class EngineLoggerHook(LoggerHook):
    """Class to log metrics with engine.
    It requires `granular-engine`_ to be installed.
    Args:
        init_kwargs (dict): A dict contains the initialization keys. Check
            https://pypi.org/project/granular-engine/ for more init arguments.
        interval (int): Logging interval (every k iterations).
            Default 10.
        ignore_last (bool): Ignore the log of last iterations in each epoch
            if less than `interval`.
            Default: True.
        reset_flag (bool): Whether to clear the output buffer after logging.
            Default: False.
        commit (bool): Save the metrics dict to the engine server and increment
            the step. If false ``engine.log`` just updates the current metrics
            dict with the row argument and metrics won't be saved until
            ``engine.log`` is called with ``commit=True``.
            Default: True.
        by_epoch (bool): Whether EpochBasedRunner is used.
            Default: True.
        log_artifact (bool): If True, artifacts in {work_dir} will be uploaded
            to engine after training ends.
            Default: True
            `New in version 1.4.3.`
        out_suffix (str or tuple[str], optional): Those filenames ending with
            ``out_suffix`` will be uploaded to engine.
            Default: ('.log.json', '.log', '.py').
            `New in version 1.4.3.`
    """

    def __init__(self,
                 init_kwargs: Optional[Dict] = None,
                 interval: int = 10,
                 ignore_last: bool = True,
                 reset_flag: bool = False,
                 commit: bool = True,
                 by_epoch: bool = True,
                 log_artifact: bool = True,
                 out_suffix: Union[str, tuple] = ('.log.json', '.log', '.py')
                 ):
        super().__init__(interval=interval, 
                         ignore_last=ignore_last, 
                         keep_local=reset_flag, 
                         log_metric_by_epoch=by_epoch)
        self.init_kwargs = init_kwargs
        self.commit = commit
        self.log_artifact = log_artifact
        self.out_suffix = out_suffix
        self.by_epoch = by_epoch
        
    def import_engine(self) -> None:
        try:
            from engine import Engine
        except ImportError:
            raise ImportError(
                'Please run "pip install granular-engine" to install engine')
        self.engine = Engine(**self.init_kwargs)

    @master_only
    def before_run(self, runner) -> None:
        super().before_run(runner)
        if self.init_kwargs:
            self.import_engine()

    @master_only
    def after_train_iter(self,
                         runner,
                         batch_idx: int,
                         data_batch: Optional[Union[dict, tuple, list]] = None,
                         outputs: Optional[dict] = None) -> None:
        """All subclasses should override this method, if they need any
        operations after each validation epoch.
        Args:
            runner (Runner): The runner of the validation process.
            metrics (Dict[str, float], optional): Evaluation results of all
                metrics on validation dataset. The keys are the names of the
                metrics, and the values are corresponding results.
        """
        if self.every_n_inner_iters(batch_idx, self.interval):
            if self.log_artifact:
                tag, log_str = runner.log_processor.get_log_after_iter(
                        runner, batch_idx, 'train')
                out = {}
                for k in tag.keys():
                    out[f"train/{k}"] = tag[k]

                self.engine.log(step=runner.iter + 1, **out)

            
    @master_only
    def after_val_epoch(self, 
                        runner, 
                        metrics: Optional[Dict[str, float]] = None) -> None:
        if self.log_artifact:
            dst_path = self.engine.meta['experimentUrl']
            os.system(f"gsutil -m cp -r {runner.work_dir}/*.py {dst_path}/ 2> /dev/null")
            os.system(f"gsutil -m cp -r {runner.work_dir}/*log* {dst_path}/ 2> /dev/null")

            best_ckpt_path = glob.glob(f"{runner.work_dir}/best*")
            if len(best_ckpt_path):
                best_ckpt_path = best_ckpt_path[0]
            else:
                best_ckpt_path = None

            out = {}
            for k in metrics.keys():
                out[f"val/{k}"] = metrics[k]

            self.engine.log(step=runner.iter + 1,
                            best=True,
                            checkpoint_path=best_ckpt_path, 
                            **out)

    @master_only
    def after_run(self, runner) -> None:
        self.engine.done()


import json, copy
import numpy as np
from tqdm import tqdm

from mmengine.registry import DATASETS
from torch.utils.data import Dataset
from typing import List, Union, Callable

from mmengine.dataset.base_dataset import Compose
from multiprocessing import Pool


@DATASETS.register_module()
class EngineCocoExt(Dataset):
    def __init__(self,
                 data_path: str,
                 raster_dir_path: str,
                 vector_dir_path: str,
                 split: str = "train",
                 class_title: str = None,
                 slice: bool = False,
                 window: tuple = None,
                 stride: tuple = None,
                 pipeline: List[Union[dict, Callable]] = [],
                 max_refetch: int = 1000,
                 ignore_index: int = 255,
                 reduce_zero_label: bool = False):
        self.raster_data_path = f"{data_path}/{raster_dir_path}"
        self.vector_data_path = f"{data_path}/{vector_dir_path}"

        self.split = split
        self.slice = slice 
        self.window = window 
        self.stride = stride
        self.class_title = class_title
        self.ignore_index = ignore_index
        self.reduce_zero_label = reduce_zero_label

        self.pipeline = Compose(pipeline)
        self.max_refetch = max_refetch

        self.get_data_list()
    
    def full_init(self):
        self.get_data_info

    def load_vector_file(self, vec_path):
        vec_map = json.load(open(vec_path))

        file_name = vec_map["images"][0]["file_name"].split("/")[-1]
        file_path = f"{self.raster_data_path}/{file_name}"

        img_map = {
            "img_path": file_path,
            "img_id": vec_map["info"]["id"],
            "height": vec_map["images"][0]["height"],
            "width": vec_map["images"][0]["width"],
            "instances": [],
            "seg_fields": [],
            "reduce_zero_label": self.reduce_zero_label
        }

        for annotation in vec_map["annotations"]:
            try:
                xmin, ymin, width, height = annotation["bbox"]
                xmax,ymax = xmin+width, ymin+height
                bbox = [xmin, ymin, xmax, ymax]
                
                bbox_label = annotation["properties"][self.class_title_idx]["labels"][0]

                assert bbox_label in range(len(self._metainfo["classes"]))
                
                mask = [annotation["segmentation"][0][:-2]]

                ann_map = {
                    "bbox": bbox,
                    "bbox_label": bbox_label,
                    "mask": mask,
                    "ignore_flag": 0,
                    "extra_anns": []
                }

                img_map["instances"] += [ann_map]
            except:
                continue

        return img_map


    def get_data_list(self):
        meta_path = f"{self.vector_data_path}/metadata.json"
        metadata = json.load(open(meta_path, "r"))

        if not self.class_title:
            self.CLASSES = metadata["label:metadata"][0]["options"]
            self.class_title = metadata["label:metadata"][0]["title"].replace(" ", "-").lower()
            self.class_title_idx = 0
        else:
            for idx, question in enumerate(metadata["label:metadata"]):
                if question["title"] == self.class_title:
                    self.CLASSES = question["options"]
                    self.class_title = self.class_title.replace(" ", "-").lower()
                    self.class_title_idx = idx 

        self._metainfo = {
            "title": metadata["title"],
            "description": metadata["description"],
            "tags": metadata["tags"],
            "problemType": metadata["problemType"] if "problemType" in metadata else None,
            "question_title": metadata["label:metadata"][self.class_title_idx]["title"],
            "question_description": metadata["label:metadata"][self.class_title_idx]["description"],
            "classes": metadata["label:metadata"][self.class_title_idx]["options"],
            "reduce_zero_label": self.reduce_zero_label
        }

        vec_path_list = [f"{self.vector_data_path}/{vec_file}" for vec_file in metadata["dataset"][self.split]]

        pool = Pool(16)
        self.data_list = list(tqdm(pool.imap(self.load_vector_file, vec_path_list), total=len(vec_path_list)))
        pool.close()

    @property
    def metainfo(self) -> dict:
        """Get meta information of dataset.

        Returns:
            dict: meta information collected from ``BaseDataset.METAINFO``,
            annotation file and metainfo argument during instantiation.
        """
        return copy.deepcopy(self._metainfo)
    
    def get_data_info(self, idx: int):
        data = self.data_list[idx]
        data["sample_idx"] = idx

        return data
    
    def prepare_data(self, idx: int):
        data = self.get_data_info(idx)
        data['dataset'] = self
        return self.pipeline(data)

    def _rand_another(self):
        return np.random.randint(0, len(self))

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx: int):
        if self.split == "test":
            data = self.prepare_data(idx)
            if not data:
                raise Exception('Test time pipline should not get `None` data_sample')

        for _ in range(self.max_refetch+1):
            data = self.prepare_data(idx)
            if not data:
                idx = self._rand_another()
                continue

            return data



from mmengine.config import Config
import importlib

def add_classes_to_package(package_name, cls_list):
    package = importlib.import_module(package_name)
    for cls in cls_list:
        class_name = cls.__name__
        
        setattr(package, class_name, cls)
        if '__all__' in package.__dict__:
            package.__dict__['__all__'].append(class_name)
        else:
            package.__dict__['__all__'] = [class_name]

def add_file_name_in_engine_tags(cfg, logger=None):
    """If engine logging is present, update tags with
        the config file name.
    """
    assert isinstance(cfg, Config), \
        f'cfg got wrong type: {type(cfg)}, expected Config'
    if 'custom_hooks' in cfg:
        assert len(cfg.custom_hooks), "Print cannot find logger hooks"
        for hook in cfg.custom_hooks:
            if hook['type'] == 'EngineLoggerHook':
                if 'init_kwargs' in hook:
                    if 'tags' in hook['init_kwargs']:
                        fname = cfg.filename.split('/')[-1].split('.')[0]
                        hook['init_kwargs']['tags'].append(fname)
    return cfg

